--- old/usr/src/man/man5/resource_controls.5	2013-07-13 17:57:07.789812986 +0000
+++ new/usr/src/man/man5/resource_controls.5	2013-07-13 17:57:07.802209314 +0000
@@ -3,7 +3,7 @@
 .\" The contents of this file are subject to the terms of the Common Development and Distribution License (the "License").  You may not use this file except in compliance with the License.
 .\" You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE or http://www.opensolaris.org/os/licensing.  See the License for the specific language governing permissions and limitations under the License.
 .\" When distributing Covered Code, include this CDDL HEADER in each file and include the License file at usr/src/OPENSOLARIS.LICENSE.  If applicable, add the following below this CDDL HEADER, with the fields enclosed by brackets "[]" replaced with your own identifying information: Portions Copyright [yyyy] [name of copyright owner]
-.TH RESOURCE_CONTROLS 5 "Jul 2, 2007"
+.TH RESOURCE_CONTROLS 5 "Jul 2, 2013"
 .SH NAME
 resource_controls \- resource controls available through project database
 .SH DESCRIPTION
@@ -171,6 +171,16 @@
 .sp
 .ne 2
 .na
+\fB\fBprocess.max-sigqueue-size\fR\fR
+.ad
+.sp .6
+.RS 4n
+Maximum number of outstanding queued signals.
+.RE
+
+.sp
+.ne 2
+.na
 \fB\fBprocess.max-stack-size\fR\fR
 .ad
 .sp .6
--- old/usr/src/uts/common/os/rctl_proc.c	2013-07-13 17:57:08.319609817 +0000
+++ new/usr/src/uts/common/os/rctl_proc.c	2013-07-13 17:57:08.330719662 +0000
@@ -23,8 +23,6 @@
  * Use is subject to license terms.
  */
 
-#pragma ident	"%Z%%M%	%I%	%E% SMI"
-
 #include <sys/types.h>
 #include <sys/cmn_err.h>
 #include <sys/sysmacros.h>
@@ -32,6 +30,7 @@
 #include <sys/rctl.h>
 #include <sys/rctl_impl.h>
 #include <sys/port_kernel.h>
+#include <sys/signal.h>
 
 #include <sys/vmparam.h>
 #include <sys/machparam.h>
@@ -65,6 +64,7 @@
 rctl_hndl_t rc_process_semmsl;
 rctl_hndl_t rc_process_semopm;
 rctl_hndl_t rc_process_portev;
+rctl_hndl_t rc_process_sigqueue;
 
 /*
  * process.max-cpu-time / RLIMIT_CPU
@@ -288,7 +288,7 @@
  *   here as the native and ILP32 infinite values for each resource control.
  */
 void
-rctlproc_init()
+rctlproc_init(void)
 {
 	rctl_set_t *set;
 	rctl_alloc_gp_t *gp;
@@ -368,6 +368,15 @@
 	rctl_add_default_limit("process.max-port-events", PORT_DEFAULT_EVENTS,
 	    RCPRIV_PRIVILEGED, RCTL_LOCAL_DENY);
 
+	rc_process_sigqueue = rctl_register("process.max-sigqueue-size",
+	    RCENTITY_PROCESS, RCTL_GLOBAL_LOWERABLE | RCTL_GLOBAL_DENY_ALWAYS |
+	    RCTL_GLOBAL_COUNT, _SIGQUEUE_SIZE_MAX, _SIGQUEUE_SIZE_MAX,
+	    &rctl_absolute_ops);
+	rctl_add_default_limit("process.max-sigqueue-size",
+	    _SIGQUEUE_SIZE_BASIC, RCPRIV_BASIC, RCTL_LOCAL_DENY);
+	rctl_add_default_limit("process.max-sigqueue-size",
+	    _SIGQUEUE_SIZE_PRIVILEGED, RCPRIV_PRIVILEGED, RCTL_LOCAL_DENY);
+
 	/*
 	 * Place minimal set of controls on "sched" process for inheritance by
 	 * processes created via newproc().
--- old/usr/src/uts/common/os/sig.c	2013-07-13 17:57:08.745693009 +0000
+++ new/usr/src/uts/common/os/sig.c	2013-07-13 17:57:08.756211913 +0000
@@ -2373,13 +2373,15 @@
 	return (0);
 }
 
-#ifndef	UCHAR_MAX
-#define	UCHAR_MAX	255
+#ifndef	INT_MAX
+#define	INT_MAX	2147483647
 #endif
 
+#define	_SIGQUEUE_PREALLOC	32	/* XXX: log scale? */
+
 /*
- * The entire pool (with maxcount entries) is pre-allocated at
- * the first sigqueue/signotify call.
+ * The pre-allocated pool (with _SIGQUEUE_PREALLOC entries) is
+ * allocated at the first sigqueue/signotify call.
  */
 sigqhdr_t *
 sigqhdralloc(size_t size, uint_t maxcount)
@@ -2388,16 +2390,16 @@
 	sigqueue_t *sq, *next;
 	sigqhdr_t *sqh;
 
-	i = (maxcount * size) + sizeof (sigqhdr_t);
-	ASSERT(maxcount <= UCHAR_MAX && i <= USHRT_MAX);
+	i = (_SIGQUEUE_PREALLOC * size) + sizeof (sigqhdr_t);
+	ASSERT(maxcount <= INT_MAX);
 	sqh = kmem_alloc(i, KM_SLEEP);
-	sqh->sqb_count = (uchar_t)maxcount;
-	sqh->sqb_maxcount = (uchar_t)maxcount;
-	sqh->sqb_size = (ushort_t)i;
+	sqh->sqb_count = maxcount;
+	sqh->sqb_maxcount = maxcount;
+	sqh->sqb_size = i;
 	sqh->sqb_pexited = 0;
 	sqh->sqb_sent = 0;
 	sqh->sqb_free = sq = (sigqueue_t *)(sqh + 1);
-	for (i = maxcount - 1; i != 0; i--) {
+	for (i = _SIGQUEUE_PREALLOC - 1; i != 0; i--) {
 		next = (sigqueue_t *)((uintptr_t)sq + size);
 		sq->sq_next = next;
 		sq = next;
@@ -2411,8 +2413,9 @@
 static void sigqrel(sigqueue_t *);
 
 /*
- * allocate a sigqueue/signotify structure from the per process
- * pre-allocated pool.
+ * Allocate a sigqueue/signotify structure from the per process
+ * pre-allocated pool or allocate a new sigqueue/signotify structure
+ * if the pre-allocated pool is exhausted.
  */
 sigqueue_t *
 sigqalloc(sigqhdr_t *sqh)
@@ -2425,12 +2428,20 @@
 		mutex_enter(&sqh->sqb_lock);
 		if (sqh->sqb_count > 0) {
 			sqh->sqb_count--;
-			sq = sqh->sqb_free;
-			sqh->sqb_free = sq->sq_next;
+			if (sqh->sqb_free == NULL) {
+				/*
+				 * The pre-allocated pool is exhausted.
+				 */
+				sq = kmem_alloc(sizeof (sigqueue_t), KM_SLEEP);
+				sq->sq_func = NULL;
+			} else {
+				sq = sqh->sqb_free;
+				sq->sq_func = sigqrel;
+				sqh->sqb_free = sq->sq_next;
+			}
 			mutex_exit(&sqh->sqb_lock);
 			bzero(&sq->sq_info, sizeof (k_siginfo_t));
 			sq->sq_backptr = sqh;
-			sq->sq_func = sigqrel;
 			sq->sq_next = NULL;
 			sq->sq_external = 0;
 		} else {
--- old/usr/src/uts/common/sys/signal.h	2013-07-13 17:57:09.226099665 +0000
+++ new/usr/src/uts/common/sys/signal.h	2013-07-13 17:57:09.235785344 +0000
@@ -300,19 +300,21 @@
 	u_longlong_t	sn_snid;	/* unique id for notification	*/
 } signotifyq_t;
 
-
 typedef struct sigqhdr {		/* sigqueue pool header		*/
 	sigqueue_t	*sqb_free;	/* free sigq struct list	*/
-	uchar_t		sqb_count;	/* sigq free count		*/
-	uchar_t		sqb_maxcount;	/* sigq max free count		*/
-	ushort_t	sqb_size;	/* size of header+free structs	*/
+	int		sqb_count;	/* sigq free count		*/
+	uint_t		sqb_maxcount;	/* sigq max free count		*/
+	size_t		sqb_size;	/* size of header+free structs	*/
 	uchar_t		sqb_pexited;	/* process has exited		*/
-	uchar_t		sqb_sent;	/* number of sigq sent		*/
+	uint_t		sqb_sent;	/* number of sigq sent		*/
 	kcondvar_t	sqb_cv;		/* waiting for a sigq struct	*/
 	kmutex_t	sqb_lock;	/* lock for sigq pool		*/
 } sigqhdr_t;
 
-#define	_SIGQUEUE_MAX	32
+#define	_SIGQUEUE_SIZE_BASIC		128	/* basic limit */
+#define	_SIGQUEUE_SIZE_PRIVILEGED	512	/* privileged limit */
+#define	_SIGQUEUE_SIZE_MAX		8192	/* maximum limit */
+
 #define	_SIGNOTIFY_MAX	32
 
 extern	void	setsigact(int, void (*)(int), const k_sigset_t *, int);
--- old/usr/src/uts/common/syscall/sigqueue.c	2013-07-13 17:57:09.609330625 +0000
+++ new/usr/src/uts/common/syscall/sigqueue.c	2013-07-13 17:57:09.623961403 +0000
@@ -26,8 +26,6 @@
 
 /* Copyright (c) 1984, 1986, 1987, 1988, 1989 AT&T */
 
-#pragma ident	"%Z%%M%	%I%	%E% SMI"
-
 #include <sys/param.h>
 #include <sys/types.h>
 #include <sys/sysmacros.h>
@@ -40,6 +38,8 @@
 #include <sys/siginfo.h>
 #include <sys/debug.h>
 
+extern rctl_hndl_t rc_process_sigqueue;
+
 static int
 sigqkill(pid_t pid, sigsend_t *sigsend)
 {
@@ -121,7 +121,7 @@
  */
 int
 sigqueue(pid_t pid, int sig, /* union sigval */ void *value,
-	int si_code, int block)
+    int si_code, int block)
 {
 	int error;
 	sigsend_t v;
@@ -133,8 +133,15 @@
 		return (set_errno(EINVAL));
 
 	if ((sqh = p->p_sigqhdr) == NULL) {
+		rlim64_t sigqsz_max;
+
+		mutex_enter(&p->p_lock);
+		sigqsz_max = rctl_enforced_value(rc_process_sigqueue,
+		    p->p_rctls, p);
+		mutex_exit(&p->p_lock);
+
 		/* Allocate sigqueue pool first time */
-		sqh = sigqhdralloc(sizeof (sigqueue_t), _SIGQUEUE_MAX);
+		sqh = sigqhdralloc(sizeof (sigqueue_t), (uint_t)sigqsz_max);
 		mutex_enter(&p->p_lock);
 		if (p->p_sigqhdr == NULL) {
 			/* hang the pool head on proc */
--- old/usr/src/uts/common/syscall/sysconfig.c	2013-07-13 17:57:10.018122401 +0000
+++ new/usr/src/uts/common/syscall/sysconfig.c	2013-07-13 17:57:10.033497392 +0000
@@ -47,6 +47,8 @@
 #include <sys/zone.h>
 #include <sys/vm_usage.h>
 
+extern rctl_hndl_t rc_process_sigqueue;
+
 long
 sysconfig(int which)
 {
@@ -140,7 +142,17 @@
 		return (_SEM_VALUE_MAX);
 
 	case _CONFIG_SIGQUEUE_MAX:
-		return (_SIGQUEUE_MAX);
+		/*
+		 * Maximum number of outstanding queued signals.
+		 */
+		{
+			rlim64_t sigqsz_max;
+			mutex_enter(&curproc->p_lock);
+			sigqsz_max = rctl_enforced_value(rc_process_sigqueue,
+			    curproc->p_rctls, curproc);
+			mutex_exit(&curproc->p_lock);
+			return ((uint_t)sigqsz_max);
+		}
 
 	case _CONFIG_SIGRT_MIN:
 		return (_SIGRTMIN);
